:- use_package([dcg,assertions]).

%----------------------------------------------------------------------
:- comment(title, "A BiBTeX parser in Prolog").
:- comment(author, "Manuel Hermenegildo").

:- comment(module, "A simple BiBTeX parser in Prolog. This is just a
   first shot, but does a reasonable job and should be easily
   extensible.").

%----------------------------------------------------------------------

bibparser(prolog_paper(AKey,[[bibtextype,LCType]|Rest])) --> 
	spaces, 
	atsign, 
	chars_no_space(Type),
	delimiter_open(X),
	spaces,
	chars_no_space(Key),
	spaces,
 	item_list(AVPL),
 	delimiter_close(X),
	spaces,
	tail_chars(_TailChars),
	{name(AKey,Key),
	 append(AVPL,[[]],Rest),
         to_lower_case(Type,LCType)}.

item_list([I|R])      --> item(I,R,RN), item_list(RN).
item_list([])         --> [].

item(Item,R,RN) --> 
	comma,
	spaces,
	chars_no_space(NameS),
	spaces,
	equal,
	spaces,
	item_body(Body),
	spaces,
	{   to_lower_case(NameS,LCNameS),
	    name(Name,LCNameS),
%% 	    format(user_error,"Name = ~w~n",[Name]),
%% 	    format(user_error,"Body = ~s~n",[Body]),
	    ( Body = "yes", 
              Item = [Name],
              R=RN
	    ; Name = author,  
	      parse_authors(Authors,Body,[]),
%%            format(user_error,"Authors = ~s~n",[Body]),
%% 	      plist(Authors),
              Item = [Name,Authors],
              R=RN
 	    ; Name = editor,  
	      parse_authors(Editors,Body,[]),
%%            format(user_error,"Editors = ~s~n",[Body]),
%%  	      plist(Editors),
              Item = [Name,Editors],
              R=RN
	    ; Name = butype, 
	      to_lower_case(Body,BibUType),
              Item = [Name,BibUType],
              R=RN
	    ; (Name = title ;  Name = booktitle),
%% No conversions...
%% 	      delete_element(Body,T1,0'{),
%% 	      delete_element(T1,NBody,0'}),
              Item = [Name,Body],
              R=RN
	    ; Name = pages,  
	      parse_pages(BP,EP,Body,[]),
              Item = [beginpage,BP],
              R = [[endpage,EP]|RN]
	    ; \+ (Name = author), 
	      \+ (Name = editor), 
	      \+ (Name = title), 
	      \+ (Name = booktitle), 
              \+ (Name = pages), 
              \+ (Name = butype), 
	      \+ (Body = "yes"),
              Item = [Name,Body],
              R=RN )
	}.


%% plist([Author]) :-
%%       format(user_error,"Parsed Author = ~s~n",[Author]).
%% plist([Author|T]) :- 
%%       format(user_error,"Parsed Author = ~s~n",[Author]),
%%       plist(T).

item_body(Body) -->
	delimiter_open(X),
	body_chars(TBody),
	delimiter_close(X),
	{prolog_escape(TBody,Body)}.
item_body(Body) -->
	chars_no_space(Body).

parse_authors([A|R]) --> 
	auth_chars(A), 
	spaces, 
	" and", 
	spaces, 
	parse_authors(R). 
parse_authors([A])   --> 
	auth_chars(A).

parse_pages(BP,EP)   --> spaces,digits(BP),digit_spacing,digits(EP),spaces.

digit_spacing  --> "-".
digit_spacing  --> "--".
digit_spacing  --> "---".

%% Now includes comments
spaces --> space, spaces.
spaces --> comm_delim, comm_chars(_), spaces.
spaces --> [].

comm_chars([H|T]) --> comm_char(H), comm_chars(T).
comm_chars([])    --> [].

comm_char(X) --> [X], {\+ eol([X],[])}.

digits([D|R])       --> digit(D), digits(R).
digits([])          --> [].

chars_no_space([H|T]) --> char_no_space(H), chars_no_space(T).
chars_no_space([])    --> [].

char_no_space(X) --> [X], 
	{\+ atsign([X],[]), \+ space([X],[]), \+ equal([X],[]), 
         \+ curly_char([X],[]), \+ comma([X],[]), \+ dbquote([X],[])}.

curly_char --> open.
curly_char --> close.

auth_chars([])    --> [].
auth_chars([H|T]) --> body_char(H), auth_chars(T).

body_chars([H|T]) --> body_char(H), body_chars(T).
body_chars([])    --> [].

body_char(X) --> [X], {\+ equal([X],[]), \+ atsign([X],[])}.

tail_chars([H|T]) --> tail_char(H), tail_chars(T).
tail_chars([])    --> [].

tail_char(X) --> [X], {\+ atsign([X],[])}.

delimiter_open(curly)  --> open.
delimiter_open(paren)  --> popen.
delimiter_open(quote)  --> dbquote.
delimiter_close(curly) --> close.
delimiter_close(paren) --> pclose.
delimiter_close(quote) --> dbquote.

dbquote    --> [0'"].
atsign     --> [0'@].
comma      --> [0',].
equal      --> [0'=].
open       --> [0'{].
popen      --> [0'(].
close      --> [0'}].
pclose     --> [0')].
digit(X)   --> [X], {X > 47, X<58}.
space      --> [X], {X=<32}. % covers EOF, ^I, ^G, ^J, ^K, ^L, ^M, ^Z, ESC, ^_
eol        --> [0'
].                         % newline (10)
comm_delim --> [0'%].

%----------------------------------------------------------------------
% simple get_bib_entry -- gets all text until next "@"
%----------------------------------------------------------------------

get_bib_entry(Entry,Terminator) :-
	get_code(Ch1),
	get_bib_entry3(Ch1, Entry, Terminator).

get_bib_entry3(Ch, [], Terminator) :-
	is_endentry(Ch),
	!,
	Terminator = Ch.
get_bib_entry3(Ch, [Ch|Line], Terminator) :-
	get_code(Next),
	get_bib_entry3(Next, Line, Terminator).

is_endentry(0'@).  % entry delimiter
is_endentry(-1).   % eof, 26 for C-Prolog and Dec10

%----------------------------------------------------------------------
:- comment(version_maintenance,dir(version)).

:- comment(version(0*4+37,2001/10/28,11:16*47+'CET'), "BiBTeX parser
   now supports comment syntax (% to end of line) (Manuel
   Hermenegildo)").

:- comment(version(0*4+20,1998/11/23,15:50*26+'MET'), "Uniquified
   multiple-arity predicates.  (Manuel Hermenegildo)").
%----------------------------------------------------------------------


